{
 "cells": [
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "1fccc8a5",
   "metadata": {},
   "source": [
    "## Classes"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "cc5b97a4",
   "metadata": {},
   "source": [
    "![](../img/class.png)"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "b4819c3a",
   "metadata": {},
   "source": [
    "### Inheritance in Python"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "15208b29",
   "metadata": {},
   "source": [
    "Have you ever had multiple classes that have similar attributes and methods? In the code below, the class `Dachshund` and `Poodle` have similar attributes (`color`) and methods (`show_info`). "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "865533b6",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "This is a Dachshund with black color.\n"
     ]
    }
   ],
   "source": [
    "class Dachshund:\n",
    "    def __init__(self, color: str):\n",
    "        self.color = color\n",
    "\n",
    "    def show_info(self):\n",
    "        print(f\"This is a Dachshund with {self.color} color.\")\n",
    "\n",
    "\n",
    "class Poodle:\n",
    "    def __init__(self, color: str):\n",
    "        self.color = color\n",
    "\n",
    "    def show_info(self):\n",
    "        print(f\"This is a Poodle with {self.color} color.\")\n",
    "\n",
    "\n",
    "bim = Dachshund(\"black\")\n",
    "bim.show_info()\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "754f0769",
   "metadata": {},
   "source": [
    "If so, use inheritance to organize your classes. Inheritance allows us to define a parent class and child classes. A child class inherits all the methods and attributes of the parent class. \n",
    "\n",
    "`super().__init__` makes the child class inherit all the methods and properties from its parent.\n",
    "\n",
    "In the code below, we define the parent class to be `Dog` and the child classes to be `Dachshund` and `Poodle`. With class inheritance, we avoid repeating the same piece of code multiple times. \n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "id": "8257ca68",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "This is a Dachshund with black color.\n"
     ]
    }
   ],
   "source": [
    "class Dog:\n",
    "    def __init__(self, type_: str, color: str):\n",
    "        self.type = type_\n",
    "        self.color = color\n",
    "\n",
    "    def show_info(self):\n",
    "        print(f\"This is a {self.type} with {self.color} color.\")\n",
    "\n",
    "\n",
    "class Dachshund(Dog):\n",
    "    def __init__(self, color: str):\n",
    "        super().__init__(type_=\"Dachshund\", color=color)\n",
    "\n",
    "\n",
    "class Poodle(Dog):\n",
    "    def __init__(self, color: str):\n",
    "        super().__init__(type_=\"Poodle\", color=color)\n",
    "\n",
    "\n",
    "bim = Dachshund(\"black\")\n",
    "bim.show_info()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "798bae91",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "This is a Poodle with brown color.\n"
     ]
    }
   ],
   "source": [
    "coco = Poodle(\"brown\")\n",
    "coco.show_info()\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "817a0eff",
   "metadata": {},
   "source": [
    "Learn more about inheritance in Python [here](https://www.w3schools.com/python/python_inheritance.asp)."
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "45c8847f",
   "metadata": {},
   "source": [
    "### Abstract Classes: Declare Methods without Implementation"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "58649bd2",
   "metadata": {},
   "source": [
    "To ensure that all subclasses implement a set of methods and properties, use abstract methods within an abstract class. This promotes code reusability and a consistent interface across different implementations.\n",
    "\n",
    "In the following code, Drink serves as an abstract class with an abstract method consume."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "3d560ac4",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:53:57.751754Z",
     "start_time": "2021-08-30T00:53:57.725292Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Drinking English Breakfast tea...\n",
      "Drinking Tropical Blast smoothie...\n"
     ]
    }
   ],
   "source": [
    "from abc import ABC, abstractmethod\n",
    "\n",
    "\n",
    "class Drink(ABC):\n",
    "    def __init__(self, name, volume):\n",
    "        self.name = name\n",
    "        self.volume = volume\n",
    "\n",
    "    @abstractmethod\n",
    "    def consume(self):\n",
    "        pass\n",
    "\n",
    "\n",
    "class Tea(Drink):\n",
    "    def consume(self):\n",
    "        print(f\"Drinking {self.name} tea...\")\n",
    "\n",
    "\n",
    "class Smoothie(Drink):\n",
    "    def consume(self):\n",
    "        print(f\"Drinking {self.name} smoothie...\")\n",
    "\n",
    "\n",
    "Tea(\"English Breakfast\", 250).consume()\n",
    "Smoothie(\"Tropical Blast\", 500).consume()"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "f2426fc4",
   "metadata": {},
   "source": [
    "### Distinguishing Instance-Level and Class Methods\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "0fd4cc9e",
   "metadata": {},
   "source": [
    "An instance-level method requires instantiating a class object to operate, while a class method doesn't.\n",
    "\n",
    "Class methods can provide alternate ways to construct objects. In the code below, the `from_csv` class method instantiates the class by reading data from a CSV file."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "id": "69f4f781",
   "metadata": {
    "tags": [
     "remove-cell"
    ]
   },
   "outputs": [],
   "source": [
    "!echo \"Name,Age,Country\" > data.csv && echo \"Alice,25,USA\" >> data.csv && echo \"Bob,30,Canada\" >> data.csv"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "id": "53c1614d",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:54:55.629438Z",
     "start_time": "2021-08-30T00:54:55.618862Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Shape of data: (3, 2)\n"
     ]
    }
   ],
   "source": [
    "import pandas as pd\n",
    "\n",
    "class DataAnalyzer:\n",
    "    def __init__(self, data):\n",
    "        self.data = data\n",
    "\n",
    "    def analyze(self): # instance-level method\n",
    "        print(f\"Shape of data: {self.data.shape}\")\n",
    "\n",
    "    @classmethod\n",
    "    def from_csv(cls, csv_path): # class method\n",
    "        data = pd.read_csv(csv_path)\n",
    "        return cls(data)\n",
    "\n",
    "\n",
    "data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})\n",
    "analyzer = DataAnalyzer(data)\n",
    "analyzer.analyze()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "id": "0afa6c77",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:55:16.233030Z",
     "start_time": "2021-08-30T00:55:16.226939Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Shape of data: (2, 3)\n"
     ]
    }
   ],
   "source": [
    "# Using the class method to create an instance from a CSV file\n",
    "csv_file_path = \"data.csv\"\n",
    "analyzer = DataAnalyzer.from_csv(csv_file_path) \n",
    "analyzer.analyze()"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "6abd916d",
   "metadata": {},
   "source": [
    "### getattr: a Better Way to Get the Attribute of a Class"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "5ab19f8e",
   "metadata": {},
   "source": [
    "If you want to get a default value when calling an attribute that is not in a class, use `getattr()` method.\n",
    "\n",
    "The `getattr(class, attribute_name)` method simply gets the value of an attribute of a class. However, if the attribute is not found in a class, it returns the default value provided to the function."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "a8a0aee1",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:59:12.972426Z",
     "start_time": "2021-08-30T00:59:12.942780Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "The color of apple is red\n"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 8;\n                var nbb_unformatted_code = \"\\nclass Food:\\n    def __init__(self, name: str, color: str):\\n        self.name = name \\n        self.color = color \\n\\napple = Food('apple', 'red')\\n\\nprint(\\\"The color of apple is\\\", getattr(apple, 'color', 'yellow'))\";\n                var nbb_formatted_code = \"class Food:\\n    def __init__(self, name: str, color: str):\\n        self.name = name\\n        self.color = color\\n\\n\\napple = Food(\\\"apple\\\", \\\"red\\\")\\n\\nprint(\\\"The color of apple is\\\", getattr(apple, \\\"color\\\", \\\"yellow\\\"))\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "class Food:\n",
    "    def __init__(self, name: str, color: str):\n",
    "        self.name = name\n",
    "        self.color = color\n",
    "\n",
    "\n",
    "apple = Food(\"apple\", \"red\")\n",
    "\n",
    "print(\"The color of apple is\", getattr(apple, \"color\", \"yellow\"))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "82d0b0d2",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:59:13.606580Z",
     "start_time": "2021-08-30T00:59:13.591580Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "The flavor of apple is sweet\n"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 9;\n                var nbb_unformatted_code = \"print(\\\"The flavor of apple is\\\", getattr(apple, 'flavor', 'sweet'))\";\n                var nbb_formatted_code = \"print(\\\"The flavor of apple is\\\", getattr(apple, \\\"flavor\\\", \\\"sweet\\\"))\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "print(\"The flavor of apple is\", getattr(apple, \"flavor\", \"sweet\"))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "8efa56b6",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:59:14.290269Z",
     "start_time": "2021-08-30T00:59:14.258917Z"
    }
   },
   "outputs": [
    {
     "ename": "AttributeError",
     "evalue": "'Food' object has no attribute 'sweet'",
     "output_type": "error",
     "traceback": [
      "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[0;31mAttributeError\u001b[0m                            Traceback (most recent call last)",
      "\u001b[0;32m/tmp/ipykernel_337430/3178150741.py\u001b[0m in \u001b[0;36m<module>\u001b[0;34m\u001b[0m\n\u001b[0;32m----> 1\u001b[0;31m \u001b[0mprint\u001b[0m\u001b[0;34m(\u001b[0m\u001b[0;34m\"The flavor of apple is\"\u001b[0m\u001b[0;34m,\u001b[0m \u001b[0mapple\u001b[0m\u001b[0;34m.\u001b[0m\u001b[0msweet\u001b[0m\u001b[0;34m)\u001b[0m\u001b[0;34m\u001b[0m\u001b[0;34m\u001b[0m\u001b[0m\n\u001b[0m",
      "\u001b[0;31mAttributeError\u001b[0m: 'Food' object has no attribute 'sweet'"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 10;\n                var nbb_unformatted_code = \"print(\\\"The flavor of apple is\\\", apple.sweet)\";\n                var nbb_formatted_code = \"print(\\\"The flavor of apple is\\\", apple.sweet)\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "print(\"The flavor of apple is\", apple.sweet)"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "4cb1dbb8",
   "metadata": {},
   "source": [
    "### `__call__`: Call your Class Instance like a Function"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "8df94fed",
   "metadata": {},
   "source": [
    "If you want to call your class instance like a function, add `__call__` method to your class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "0f79c0b1",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T00:59:18.557329Z",
     "start_time": "2021-08-30T00:59:18.532123Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Instance is created\n",
      "Instance is called\n"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 11;\n                var nbb_unformatted_code = \"class DataLoader:\\n\\n\\tdef __init__(self, data_dir: str):\\n\\t\\tself.data_dir = data_dir\\n\\t\\tprint(\\\"Instance is created\\\")\\n\\n\\tdef __call__(self):\\n\\t\\tprint(\\\"Instance is called\\\")\\n\\ndata_loader = DataLoader('my_data_dir')\\n# Instance is created\\n\\ndata_loader()\\n# Instance is called\";\n                var nbb_formatted_code = \"class DataLoader:\\n    def __init__(self, data_dir: str):\\n        self.data_dir = data_dir\\n        print(\\\"Instance is created\\\")\\n\\n    def __call__(self):\\n        print(\\\"Instance is called\\\")\\n\\n\\ndata_loader = DataLoader(\\\"my_data_dir\\\")\\n# Instance is created\\n\\ndata_loader()\\n# Instance is called\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "class DataLoader:\n",
    "    def __init__(self, data_dir: str):\n",
    "        self.data_dir = data_dir\n",
    "        print(\"Instance is created\")\n",
    "\n",
    "    def __call__(self):\n",
    "        print(\"Instance is called\")\n",
    "\n",
    "\n",
    "data_loader = DataLoader(\"my_data_dir\")\n",
    "# Instance is created\n",
    "\n",
    "data_loader()\n",
    "# Instance is called"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "eed13008",
   "metadata": {},
   "source": [
    "### Instance Comparison in Python Classes"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "0d605ea9",
   "metadata": {},
   "source": [
    "Even if two class instances have the same attributes, they are not equal because they are stored in separate memory locations.\n",
    "\n",
    "To define how class instances should be compared, add the `__eq__` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "d61ad733",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "False"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "class Dog:\n",
    "    def __init__(self, name: str):\n",
    "        self.name = name\n",
    "\n",
    "\n",
    "dog1 = Dog(\"Bim\")\n",
    "dog2 = Dog(\"Bim\")\n",
    "dog1 == dog2\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "ffba882d",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "True"
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "class Dog:\n",
    "    def __init__(self, name: str):\n",
    "        self.name = name\n",
    "\n",
    "    def __eq__(self, other):\n",
    "        return self.name == other.name\n",
    "\n",
    "\n",
    "dog1 = Dog(\"Bim\")\n",
    "dog2 = Dog(\"Bim\")\n",
    "dog1 == dog2\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "7c083eeb",
   "metadata": {},
   "source": [
    "### Static method: use the function without adding the attributes required for a new instance"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "6218e4de",
   "metadata": {},
   "source": [
    "Have you ever had a function in your class that doesn’t access any properties of a class but fits well in a class? You might find it redundant to instantiate the class to use that function. That is when you can turn your function into a static method.\n",
    "\n",
    "All you need to turn your function into a static method is the decorator `@staticmethod`. Now you can use the function without adding the attributes required for a new instance.\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "3b7208d6",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T01:00:17.808040Z",
     "start_time": "2021-08-30T01:00:17.775913Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "My favorite page is \n"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 12;\n                var nbb_unformatted_code = \"import re \\n\\nclass ProcessText:\\n    def __init__(self, text_column: str):\\n        self.text_column = text_column\\n\\n    @staticmethod\\n    def remove_URL(sample: str) -> str:\\n        \\\"\\\"\\\"Replace url with empty space\\\"\\\"\\\"\\n        return re.sub(r'http\\\\S+', \\\"\\\", sample)\\n\\ntext = ProcessText.remove_URL('My favorite page is https://www.google.com')\\nprint(text)\";\n                var nbb_formatted_code = \"import re\\n\\n\\nclass ProcessText:\\n    def __init__(self, text_column: str):\\n        self.text_column = text_column\\n\\n    @staticmethod\\n    def remove_URL(sample: str) -> str:\\n        \\\"\\\"\\\"Replace url with empty space\\\"\\\"\\\"\\n        return re.sub(r\\\"http\\\\S+\\\", \\\"\\\", sample)\\n\\n\\ntext = ProcessText.remove_URL(\\\"My favorite page is https://www.google.com\\\")\\nprint(text)\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "import re\n",
    "\n",
    "\n",
    "class ProcessText:\n",
    "    def __init__(self, text_column: str):\n",
    "        self.text_column = text_column\n",
    "\n",
    "    @staticmethod\n",
    "    def remove_URL(sample: str) -> str:\n",
    "        \"\"\"Replace url with empty space\"\"\"\n",
    "        return re.sub(r\"http\\S+\", \"\", sample)\n",
    "\n",
    "\n",
    "text = ProcessText.remove_URL(\"My favorite page is https://www.google.com\")\n",
    "print(text)"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "2a101e4d",
   "metadata": {},
   "source": [
    "### Minimize Data Risks with Python Private Variables"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "8093f7d3",
   "metadata": {},
   "source": [
    "To restrict external access and modification of a variable outside of a class, make it a private variable by using double underscores. This helps minimize the chances of unintended alterations."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "65e3e68a",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "The price of Apples is $2.99\n"
     ]
    },
    {
     "ename": "AttributeError",
     "evalue": "'Grocery' object has no attribute '__price'",
     "output_type": "error",
     "traceback": [
      "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[0;31mAttributeError\u001b[0m                            Traceback (most recent call last)",
      "Cell \u001b[0;32mIn[4], line 18\u001b[0m\n\u001b[1;32m     15\u001b[0m grocery_item\u001b[39m.\u001b[39mget_price()\n\u001b[1;32m     17\u001b[0m \u001b[39m# Access the private variable directly\u001b[39;00m\n\u001b[0;32m---> 18\u001b[0m grocery_item\u001b[39m.\u001b[39;49m__price\n",
      "\u001b[0;31mAttributeError\u001b[0m: 'Grocery' object has no attribute '__price'"
     ]
    }
   ],
   "source": [
    "class Grocery:\n",
    "    def __init__(self, item, price):\n",
    "        # Making 'price' a private variable\n",
    "        self.__price = price\n",
    "        self.item = item\n",
    "\n",
    "    # Getter method to access the private variable 'price'\n",
    "    def get_price(self):\n",
    "        print(f\"The price of {self.item} is ${self.__price}\")\n",
    "\n",
    "\n",
    "# Create an instance of the Grocery class\n",
    "grocery_item = Grocery(\"Apples\", 2.99)\n",
    "\n",
    "# Access the private variable 'price' using the getter method\n",
    "grocery_item.get_price()\n",
    "\n",
    "# Access the private variable directly\n",
    "grocery_item.__price\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "2b37956a",
   "metadata": {},
   "source": [
    "### Ensure Data Integrity with Getters and Setters in Python"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "1188b89a",
   "metadata": {},
   "source": [
    "In Python, the property decorator controls property access and modification through getters and setters.\n",
    "\n",
    "For example, in a `BankAccount` class, without getters and setters, the balance can be directly modified, potentially leading to an invalid state."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "c3d5a814",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "100\n",
      "-50\n"
     ]
    }
   ],
   "source": [
    "class BankAccount:\n",
    "    def __init__(self, initial_balance: float):\n",
    "        self.balance = initial_balance\n",
    "\n",
    "account = BankAccount(100)\n",
    "print(account.balance) \n",
    "\n",
    "# Directly modifying the balance (no validation)\n",
    "account.balance = -50\n",
    "print(account.balance) "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8843dd71",
   "metadata": {},
   "source": [
    "Using getters and setters ensures the balance cannot be set to an invalid value."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "ce4529d1",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "100\n"
     ]
    },
    {
     "ename": "ValueError",
     "evalue": "Balance cannot be negative",
     "output_type": "error",
     "traceback": [
      "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[0;31mValueError\u001b[0m                                Traceback (most recent call last)",
      "Cell \u001b[0;32mIn[6], line 21\u001b[0m\n\u001b[1;32m     18\u001b[0m \u001b[38;5;28mprint\u001b[39m(account\u001b[38;5;241m.\u001b[39mbalance) \n\u001b[1;32m     20\u001b[0m \u001b[38;5;66;03m# Attempting to set an invalid balance\u001b[39;00m\n\u001b[0;32m---> 21\u001b[0m \u001b[43maccount\u001b[49m\u001b[38;5;241;43m.\u001b[39;49m\u001b[43mbalance\u001b[49m \u001b[38;5;241m=\u001b[39m \u001b[38;5;241m-\u001b[39m\u001b[38;5;241m50\u001b[39m \n",
      "Cell \u001b[0;32mIn[6], line 12\u001b[0m, in \u001b[0;36mBankAccount.balance\u001b[0;34m(self, value)\u001b[0m\n\u001b[1;32m      9\u001b[0m \u001b[38;5;129m@balance\u001b[39m\u001b[38;5;241m.\u001b[39msetter\n\u001b[1;32m     10\u001b[0m \u001b[38;5;28;01mdef\u001b[39;00m \u001b[38;5;21mbalance\u001b[39m(\u001b[38;5;28mself\u001b[39m, value):\n\u001b[1;32m     11\u001b[0m     \u001b[38;5;28;01mif\u001b[39;00m value \u001b[38;5;241m<\u001b[39m \u001b[38;5;241m0\u001b[39m:\n\u001b[0;32m---> 12\u001b[0m         \u001b[38;5;28;01mraise\u001b[39;00m \u001b[38;5;167;01mValueError\u001b[39;00m(\u001b[38;5;124m\"\u001b[39m\u001b[38;5;124mBalance cannot be negative\u001b[39m\u001b[38;5;124m\"\u001b[39m)\n\u001b[1;32m     13\u001b[0m     \u001b[38;5;28mself\u001b[39m\u001b[38;5;241m.\u001b[39m_balance \u001b[38;5;241m=\u001b[39m value\n",
      "\u001b[0;31mValueError\u001b[0m: Balance cannot be negative"
     ]
    }
   ],
   "source": [
    "class BankAccount:\n",
    "    def __init__(self, initial_balance: float):\n",
    "        self._balance = initial_balance\n",
    "\n",
    "    @property\n",
    "    def balance(self):\n",
    "        return self._balance\n",
    "\n",
    "    @balance.setter\n",
    "    def balance(self, value):\n",
    "        if value < 0:\n",
    "            raise ValueError(\"Balance cannot be negative\")\n",
    "        self._balance = value\n",
    "\n",
    "\n",
    "# Example usage\n",
    "account = BankAccount(100)\n",
    "print(account.balance) \n",
    "\n",
    "# Attempting to set an invalid balance\n",
    "account.balance = -50 "
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "c510dd76",
   "metadata": {},
   "source": [
    "### `__str__` and `__repr__`: Create a String Representation of a Python Object\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "82e6f70e",
   "metadata": {},
   "source": [
    "If you want to create a string representation of an object, add `__str__` and `__repr__`.\n",
    "\n",
    "`__str__` shows readable outputs when printing the object. `__repr__` shows outputs that are useful for displaying and debugging the object."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "id": "71450d29",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T01:45:12.288927Z",
     "start_time": "2021-08-30T01:45:12.261626Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "red apple\n"
     ]
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 17;\n                var nbb_unformatted_code = \"class Food:\\n    def __init__(self, name: str, color: str):\\n        self.name = name\\n        self.color = color\\n\\n    def __str__(self):\\n        return f\\\"{self.color} {self.name}\\\"\\n\\n    def __repr__(self):\\n        return f\\\"Food({self.color}, {self.name})\\\"\\n\\n\\nfood = Food(\\\"apple\\\", \\\"red\\\")\\n\\nprint(food)  #  str__\";\n                var nbb_formatted_code = \"class Food:\\n    def __init__(self, name: str, color: str):\\n        self.name = name\\n        self.color = color\\n\\n    def __str__(self):\\n        return f\\\"{self.color} {self.name}\\\"\\n\\n    def __repr__(self):\\n        return f\\\"Food({self.color}, {self.name})\\\"\\n\\n\\nfood = Food(\\\"apple\\\", \\\"red\\\")\\n\\nprint(food)  #  str__\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "class Food:\n",
    "    def __init__(self, name: str, color: str):\n",
    "        self.name = name\n",
    "        self.color = color\n",
    "\n",
    "    def __str__(self):\n",
    "        return f\"{self.color} {self.name}\"\n",
    "\n",
    "    def __repr__(self):\n",
    "        return f\"Food({self.color}, {self.name})\"\n",
    "\n",
    "\n",
    "food = Food(\"apple\", \"red\")\n",
    "\n",
    "print(food)  #  str__"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "id": "e6cd59a2",
   "metadata": {
    "ExecuteTime": {
     "end_time": "2021-08-30T01:45:12.705910Z",
     "start_time": "2021-08-30T01:45:12.692056Z"
    }
   },
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Food(red, apple)"
      ]
     },
     "execution_count": 18,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "data": {
      "application/javascript": "\n            setTimeout(function() {\n                var nbb_cell_id = 18;\n                var nbb_unformatted_code = \"food  # __repr__\";\n                var nbb_formatted_code = \"food  # __repr__\";\n                var nbb_cells = Jupyter.notebook.get_cells();\n                for (var i = 0; i < nbb_cells.length; ++i) {\n                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n                             nbb_cells[i].set_text(nbb_formatted_code);\n                        }\n                        break;\n                    }\n                }\n            }, 500);\n            ",
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "food  # __repr__"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "df12dd1f",
   "metadata": {},
   "source": [
    "### `__add__`: Add the Attributes of Two Class Instances"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "718af7c3",
   "metadata": {},
   "source": [
    "If you want to add the attributes of class instances, use `__add__`. In the code below, I use `__add__` to add the ages of two class instances `bim` and `coco` when calling `bim + coco`. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "f455c75a",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "7"
      ]
     },
     "execution_count": 1,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "class Dog:\n",
    "    def __init__(self, age: int):\n",
    "        self.age = age\n",
    "\n",
    "    def __add__(self, other):\n",
    "        return self.age + other.age\n",
    "\n",
    "\n",
    "class Cat:\n",
    "    def __init__(self, age: int):\n",
    "        self.age = age\n",
    "\n",
    "    def __add__(self, other):\n",
    "        return self.age + other.age\n",
    "\n",
    "\n",
    "bim = Dog(age=5)\n",
    "coco = Cat(age=2)\n",
    "bim + coco\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "cbce2b65",
   "metadata": {},
   "source": [
    "### Optimizing Memory Usage in Python with Slots"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "dcc6d440",
   "metadata": {
    "tags": [
     "hide-cell"
    ]
   },
   "outputs": [],
   "source": [
    "!pip install memory_profiler"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "847b4a77",
   "metadata": {},
   "source": [
    "In Python, objects can store their attributes in a flexible dictionary-like structure that can use a lot of memory. Slots make your objects more memory-efficient by reserving space for their attributes ahead of time.\n",
    "\n",
    "The code below shows that using slots significant reduces the memory usage."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c3b401b5",
   "metadata": {},
   "outputs": [],
   "source": [
    "%%writefile without_slot.py\n",
    "from random import randint\n",
    "from memory_profiler import profile\n",
    "\n",
    "\n",
    "class Dog:\n",
    "    def __init__(self, age):\n",
    "        self.age = age\n",
    "\n",
    "\n",
    "@profile\n",
    "def main():\n",
    "    return [Dog(age=randint(0, 30)) for _ in range(100000)]\n",
    "\n",
    "\n",
    "if __name__ == \"__main__\":\n",
    "    main()"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "d9225e6a",
   "metadata": {},
   "source": [
    "```bash\n",
    "$ python -m memory_profiler without_slot.py\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "313fa5e9",
   "metadata": {
    "tags": [
     "remove-input"
    ]
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Filename: without_slot.py\n",
      "\n",
      "Line #    Mem usage    Increment  Occurrences   Line Contents\n",
      "=============================================================\n",
      "    10     41.6 MiB     41.6 MiB           1   @profile\n",
      "    11                                         def main():\n",
      "    12     57.8 MiB     16.2 MiB      100003       return [Dog(randint(0, 30)) for _ in range(100000)]\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "!python -m memory_profiler without_slot.py"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a0d28226",
   "metadata": {},
   "outputs": [],
   "source": [
    "%%writefile with_slot.py\n",
    "from random import randint\n",
    "from memory_profiler import profile\n",
    "\n",
    "\n",
    "class Dog:\n",
    "    # defining slots\n",
    "    __slots__ = [\"age\"] \n",
    "    def __init__(self, age):\n",
    "        self.age = age\n",
    "\n",
    "\n",
    "@profile\n",
    "def main():\n",
    "    return [Dog(age=randint(0, 30)) for _ in range(100000)]\n",
    "\n",
    "\n",
    "if __name__ == \"__main__\":\n",
    "    main()"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "id": "6466485d",
   "metadata": {},
   "source": [
    "```bash\n",
    "$ python -m memory_profiler with_slot.py\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "id": "cdd60556",
   "metadata": {
    "tags": [
     "remove-input"
    ]
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Filename: with_slot.py\n",
      "\n",
      "Line #    Mem usage    Increment  Occurrences   Line Contents\n",
      "=============================================================\n",
      "    11     41.3 MiB     41.3 MiB           1   @profile\n",
      "    12                                         def main():\n",
      "    13     46.7 MiB      5.4 MiB      100003       return [Dog(randint(0, 30)) for _ in range(100000)]\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "!python -m memory_profiler with_slot.py"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Improve Code Readability with Enums"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Hard-coded values without proper context can decrease code readability.\n",
    "\n",
    "By using an Enum, meaningful names can be assigned to these values, improving code readability."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "You can go through the gate.\n"
     ]
    }
   ],
   "source": [
    "current_status = 200\n",
    "if current_status == 200:\n",
    "    print(\"You can go through the gate.\")\n",
    "elif current_status == 500:\n",
    "    print(\"You can't go through the gate.\")\n",
    "else:\n",
    "    print(\"Invalid status code.\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "You can go through the gate.\n"
     ]
    }
   ],
   "source": [
    "from enum import Enum\n",
    "\n",
    "\n",
    "class StatusCode(Enum):\n",
    "    OK = 200\n",
    "    ERROR = 500\n",
    "\n",
    "\n",
    "current_status = 200\n",
    "if current_status == StatusCode.OK.value:\n",
    "    print(\"You can go through the gate.\")\n",
    "elif current_status == StatusCode.ERROR.value:\n",
    "    print(\"You can't go through the gate.\")\n",
    "else:\n",
    "    print(\"Invalid status code.\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Learn more about Enum](https://docs.python.org/3/library/enum.html)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2d9d81c1",
   "metadata": {},
   "source": [
    "### Embrace the Open-Closed Principle to Design Extensible Classes"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "78fd8d08",
   "metadata": {},
   "source": [
    "You should design classes that are open for extension but closed for modification (Open-Closed Principle).\n",
    "\n",
    "In the current implementation of the `DataPipeline` class, the data processing methods are directly implemented within the class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "3521c193",
   "metadata": {},
   "outputs": [],
   "source": [
    "import pandas as pd\n",
    "\n",
    "\n",
    "class DataPipeline:\n",
    "    def drop_missing_data(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return data.dropna()\n",
    "\n",
    "    def standardize_data(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return (data - data.mean()) / data.std()\n",
    "\n",
    "    def process(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return data.pipe(self.drop_missing_data).pipe(self.standardize_data)\n",
    "\n",
    "\n",
    "pipeline = DataPipeline()\n",
    "data = pd.DataFrame({\"A\": [1, 2, 3, None, 5], \"B\": [5, 4, 2, 1, 3]})\n",
    "processed_data = pipeline.process(data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ade0471c",
   "metadata": {},
   "source": [
    "Adding another processing step to the pipeline requires modifying the `process` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "801bc2be",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>A</th>\n",
       "      <th>B</th>\n",
       "      <th>C_a</th>\n",
       "      <th>C_b</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>-1.024695</td>\n",
       "      <td>1.161895</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>-0.439155</td>\n",
       "      <td>0.387298</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>0.146385</td>\n",
       "      <td>-1.161895</td>\n",
       "      <td>-1.5</td>\n",
       "      <td>1.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>1.317465</td>\n",
       "      <td>-0.387298</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "          A         B  C_a  C_b\n",
       "0 -1.024695  1.161895  0.5 -0.5\n",
       "1 -0.439155  0.387298  0.5 -0.5\n",
       "2  0.146385 -1.161895 -1.5  1.5\n",
       "4  1.317465 -0.387298  0.5 -0.5"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "import pandas as pd\n",
    "\n",
    "\n",
    "class DataPipeline:\n",
    "    def drop_missing_data(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return data.dropna()\n",
    "\n",
    "    def standardize_data(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return (data - data.mean()) / data.std()\n",
    "\n",
    "    # Adding this\n",
    "    def encode_categorical_data(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return pd.get_dummies(data)\n",
    "\n",
    "    # Requires modifying the code\n",
    "    def process(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return (\n",
    "            data.pipe(self.drop_missing_data)\n",
    "            .pipe(self.encode_categorical_data)\n",
    "            .pipe(self.standardize_data)\n",
    "        )\n",
    "\n",
    "\n",
    "pipeline = DataPipeline()\n",
    "data = pd.DataFrame(\n",
    "    {\"A\": [1, 2, 3, None, 5], \"B\": [5, 4, 2, 1, 3], \"C\": [\"a\", \"a\", \"b\", \"b\", \"a\"]}\n",
    ")\n",
    "pipeline.process(data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aae64bef",
   "metadata": {},
   "source": [
    "Refactor the `DataPipeline` class to accept pluggable strategies."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "2fcee4aa",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>A</th>\n",
       "      <th>B</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>-1.024695</td>\n",
       "      <td>1.161895</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>-0.439155</td>\n",
       "      <td>0.387298</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>0.146385</td>\n",
       "      <td>-1.161895</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>1.317465</td>\n",
       "      <td>-0.387298</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "          A         B\n",
       "0 -1.024695  1.161895\n",
       "1 -0.439155  0.387298\n",
       "2  0.146385 -1.161895\n",
       "4  1.317465 -0.387298"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "from abc import ABC, abstractmethod\n",
    "import pandas as pd\n",
    "\n",
    "\n",
    "class DataProcessingStrategy(ABC):\n",
    "    @abstractmethod\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        pass\n",
    "\n",
    "\n",
    "class DropMissingDataStrategy(DataProcessingStrategy):\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return data.dropna()\n",
    "\n",
    "\n",
    "class StandardizeDataStrategy(DataProcessingStrategy):\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return (data - data.mean()) / data.std()\n",
    "\n",
    "\n",
    "class DataPipeline:\n",
    "    def __init__(self):\n",
    "        self.strategies = []\n",
    "\n",
    "    def add_strategy(self, strategy: DataProcessingStrategy):\n",
    "        self.strategies.append(strategy)\n",
    "\n",
    "    def process(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        for strategy in self.strategies:\n",
    "            data = strategy.apply(data)\n",
    "        return data\n",
    "\n",
    "\n",
    "pipeline = DataPipeline()\n",
    "pipeline.add_strategy(DropMissingDataStrategy())\n",
    "pipeline.add_strategy(StandardizeDataStrategy())\n",
    "\n",
    "# Imagine we have some sample data\n",
    "data = pd.DataFrame({\"A\": [1, 2, 3, None, 5], \"B\": [5, 4, 2, 1, 3]})\n",
    "\n",
    "pipeline.process(data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "daea77ec",
   "metadata": {},
   "source": [
    "This design ensures the pipeline's processing strategies can be swapped, extended, or reordered without modifying the `DataPipeline` class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "5cf06e50",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>A</th>\n",
       "      <th>B</th>\n",
       "      <th>C_a</th>\n",
       "      <th>C_b</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>-1.024695</td>\n",
       "      <td>1.161895</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>-0.439155</td>\n",
       "      <td>0.387298</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>0.146385</td>\n",
       "      <td>-1.161895</td>\n",
       "      <td>-1.5</td>\n",
       "      <td>1.5</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>1.317465</td>\n",
       "      <td>-0.387298</td>\n",
       "      <td>0.5</td>\n",
       "      <td>-0.5</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "          A         B  C_a  C_b\n",
       "0 -1.024695  1.161895  0.5 -0.5\n",
       "1 -0.439155  0.387298  0.5 -0.5\n",
       "2  0.146385 -1.161895 -1.5  1.5\n",
       "4  1.317465 -0.387298  0.5 -0.5"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "from abc import ABC, abstractmethod\n",
    "import pandas as pd\n",
    "\n",
    "\n",
    "class DataProcessingStrategy(ABC):\n",
    "    @abstractmethod\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        pass\n",
    "\n",
    "\n",
    "class DropMissingDataStrategy(DataProcessingStrategy):\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return data.dropna()\n",
    "\n",
    "\n",
    "class StandardizeDataStrategy(DataProcessingStrategy):\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return (data - data.mean()) / data.std()\n",
    "\n",
    "\n",
    "class EncodeDataStrategy(DataProcessingStrategy):\n",
    "    def apply(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        return pd.get_dummies(data)\n",
    "\n",
    "\n",
    "class DataPipeline:\n",
    "    def __init__(self):\n",
    "        self.strategies = []\n",
    "\n",
    "    def add_strategy(self, strategy: DataProcessingStrategy):\n",
    "        self.strategies.append(strategy)\n",
    "\n",
    "    def process(self, data: pd.DataFrame) -> pd.DataFrame:\n",
    "        for strategy in self.strategies:\n",
    "            data = strategy.apply(data)\n",
    "        return data\n",
    "\n",
    "\n",
    "pipeline = DataPipeline()\n",
    "pipeline.add_strategy(DropMissingDataStrategy())\n",
    "pipeline.add_strategy(EncodeDataStrategy())\n",
    "pipeline.add_strategy(StandardizeDataStrategy())\n",
    "\n",
    "data = pd.DataFrame(\n",
    "    {\"A\": [1, 2, 3, None, 5], \"B\": [5, 4, 2, 1, 3], \"C\": [\"a\", \"a\", \"b\", \"b\", \"a\"]}\n",
    ")\n",
    "\n",
    "pipeline.process(data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3fb5898e",
   "metadata": {},
   "source": [
    "### Use Mixins Over Inheritance for Enhanced Modularity"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "520deabb",
   "metadata": {},
   "source": [
    "Use mixin instead of inheritance to add shared functionality without changing the primary structure of a class.\n",
    "\n",
    "In the following example, the inheritance-based approach assumes that the `PickleableModel` class has a `.model` attribute that needs to be serialized. This assumption may not be true for all subclasses."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "id": "633f0fbe",
   "metadata": {},
   "outputs": [],
   "source": [
    "import joblib\n",
    "from sklearn.cluster import KMeans\n",
    "from sklearn.svm import SVC\n",
    "from sklearn.datasets import make_blobs\n",
    "\n",
    "\n",
    "class PickleableModel:\n",
    "    def to_pickle(self, file_path):\n",
    "        \"\"\"\n",
    "        Serialize the object to a pickle file.\n",
    "        \"\"\"\n",
    "        with open(file_path, \"wb\") as file:\n",
    "            # It's important to only serialize the model attribute here\n",
    "            # otherwise joblib might fail to serialize the object\n",
    "            joblib.dump(self.model, file)\n",
    "\n",
    "    @classmethod\n",
    "    def from_pickle(cls, file_path):\n",
    "        \"\"\"\n",
    "        Deserialize pickle file to an object.\n",
    "        \"\"\"\n",
    "        with open(file_path, \"rb\") as file:\n",
    "            obj = cls()\n",
    "            obj.model = joblib.load(file)\n",
    "            return obj\n",
    "\n",
    "\n",
    "class PickleableKmeans(PickleableModel):\n",
    "    def __init__(self, n_clusters=3, **kwargs):\n",
    "        self.model = KMeans(n_clusters=n_clusters, **kwargs)\n",
    "\n",
    "    def fit(self, X, y=None):\n",
    "        self.model.fit(X)\n",
    "\n",
    "    def predict(self, X):\n",
    "        return self.model.predict(X)\n",
    "\n",
    "\n",
    "class PickleableSVM(PickleableModel):\n",
    "    def __init__(self, C=1.0, kernel=\"rbf\", **kwargs):\n",
    "        self.model = SVC(C=C, kernel=kernel, **kwargs)\n",
    "\n",
    "    def fit(self, X, y):\n",
    "        self.model.fit(X, y)\n",
    "\n",
    "    def predict(self, X):\n",
    "        return self.model.predict(X)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "id": "948892b0",
   "metadata": {},
   "outputs": [],
   "source": [
    "X, _ = make_blobs(n_samples=300, centers=4, n_features=2, random_state=42)\n",
    "\n",
    "kmeans = PickleableKmeans(n_clusters=3, n_init=\"auto\")\n",
    "kmeans.fit(X)\n",
    "\n",
    "# Serialize the model to a file\n",
    "kmeans_file_path = \"kmeans_model.pkl\"\n",
    "kmeans.to_pickle(kmeans_file_path)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4d3461d7",
   "metadata": {},
   "source": [
    "The `PickleableMixin` can be applied to any class that needs serialization, not just machine learning models."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "id": "67cb0651",
   "metadata": {},
   "outputs": [],
   "source": [
    "class PickleableMixin:\n",
    "    def to_pickle(self, file_path):\n",
    "        \"\"\"\n",
    "        Serialize the object to a pickle file.\n",
    "        \"\"\"\n",
    "        with open(file_path, \"wb\") as file:\n",
    "            joblib.dump(self, file)\n",
    "\n",
    "    @staticmethod\n",
    "    def from_pickle(file_path):\n",
    "        \"\"\"\n",
    "        Deserialize pickle file to an object.\n",
    "        \"\"\"\n",
    "        with open(file_path, \"rb\") as file:\n",
    "            return joblib.load(file)\n",
    "\n",
    "\n",
    "# Enhanced models with serialization capability\n",
    "class PickleableKmeans(KMeans, PickleableMixin):\n",
    "    pass\n",
    "\n",
    "\n",
    "class PickleableSVM(SVC, PickleableMixin):\n",
    "    pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "id": "ac14121d",
   "metadata": {},
   "outputs": [],
   "source": [
    "kmeans = PickleableKmeans(n_clusters=3, n_init=\"auto\")\n",
    "kmeans.fit(X)\n",
    "\n",
    "# Serialize the model to a file\n",
    "kmeans_file_path = \"kmeans_model.pkl\"\n",
    "kmeans.to_pickle(kmeans_file_path)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "647169d6",
   "metadata": {},
   "source": [
    "### Embracing Duck Typing for Cleaner, More Adaptable Data Science Code"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6fce343a",
   "metadata": {},
   "source": [
    "Duck typing comes from the phrase \"If it walks like a duck and quacks like a duck, then it must be a duck.\" \n",
    "\n",
    "This lets you write flexible code that works with different types of objects, as long as they have the methods or attributes you're using.\n",
    "\n",
    "For data scientists, duck typing allows creating versatile functions that work with various data structures without explicitly checking their types.\n",
    "\n",
    "Here's a simple example:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "26fe0dfe",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.5811388300841898\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n"
     ]
    }
   ],
   "source": [
    "import numpy as np\n",
    "import pandas as pd\n",
    "\n",
    "\n",
    "class CustomDataFrame:\n",
    "    def __init__(self, data):\n",
    "        self.data = data\n",
    "\n",
    "    def mean(self):\n",
    "        return np.mean(self.data)\n",
    "\n",
    "    def std(self):\n",
    "        return np.std(self.data)\n",
    "\n",
    "\n",
    "def analyze_data(data):\n",
    "    print(f\"Mean: {data.mean()}\")\n",
    "    print(f\"Standard Deviation: {data.std()}\")\n",
    "\n",
    "\n",
    "# These all work, thanks to duck typing\n",
    "numpy_array = np.array([1, 2, 3, 4, 5])\n",
    "pandas_series = pd.Series([1, 2, 3, 4, 5])\n",
    "custom_df = CustomDataFrame([1, 2, 3, 4, 5])\n",
    "\n",
    "analyze_data(numpy_array)\n",
    "analyze_data(pandas_series)\n",
    "analyze_data(custom_df)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "96171599",
   "metadata": {},
   "source": [
    "In this example, the `analyze_data` function works with NumPy arrays, Pandas Series, and our custom `CustomDataFrame` class, because they all have `mean` and `std` methods. This flexibility is powerful in data science workflows where you might be working with various data structures.\n",
    "\n",
    "This flexibility is valuable in data science because:\n",
    "\n",
    "1. It saves time: You don't need separate functions for different data types.\n",
    "\n",
    "Bad example:\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "b84455c8",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.5811388300841898\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n"
     ]
    }
   ],
   "source": [
    "def analyze_numpy_array(data):\n",
    "    print(f\"Mean: {np.mean(data)}\")\n",
    "    print(f\"Standard Deviation: {np.std(data)}\")\n",
    "\n",
    "def analyze_pandas_series(data):\n",
    "    print(f\"Mean: {data.mean()}\")\n",
    "    print(f\"Standard Deviation: {data.std()}\")\n",
    "\n",
    "def analyze_custom_df(data):\n",
    "    print(f\"Mean: {data.mean()}\")\n",
    "    print(f\"Standard Deviation: {data.std()}\")\n",
    "\n",
    "numpy_array = np.array([1, 2, 3, 4, 5])\n",
    "pandas_series = pd.Series([1, 2, 3, 4, 5])\n",
    "custom_df = CustomDataFrame([1, 2, 3, 4, 5])\n",
    "\n",
    "analyze_numpy_array(numpy_array)\n",
    "analyze_pandas_series(pandas_series)\n",
    "analyze_custom_df(custom_df)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c569cd67",
   "metadata": {},
   "source": [
    "2. It's cleaner: You avoid lots of `if` statements checking for types."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "e463b7aa",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.5811388300841898\n",
      "Mean: 3.0\n",
      "Standard Deviation: 1.4142135623730951\n"
     ]
    }
   ],
   "source": [
    "def analyze_data(data):\n",
    "    if isinstance(data, np.ndarray):\n",
    "        mean = np.mean(data)\n",
    "        std = np.std(data)\n",
    "    elif isinstance(data, pd.Series):\n",
    "        mean = data.mean()\n",
    "        std = data.std()\n",
    "    elif isinstance(data, CustomDataFrame):\n",
    "        mean = data.mean()\n",
    "        std = data.std()\n",
    "    else:\n",
    "        raise TypeError(\"Unsupported data type\")\n",
    "    \n",
    "    print(f\"Mean: {mean}\")\n",
    "    print(f\"Standard Deviation: {std}\")\n",
    "\n",
    "numpy_array = np.array([1, 2, 3, 4, 5])\n",
    "pandas_series = pd.Series([1, 2, 3, 4, 5])\n",
    "custom_df = CustomDataFrame([1, 2, 3, 4, 5])\n",
    "python_list = [1, 2, 3, 4, 5]\n",
    "\n",
    "analyze_data(numpy_array)\n",
    "analyze_data(pandas_series)\n",
    "analyze_data(custom_df)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4591a347",
   "metadata": {},
   "source": [
    "3. It's more adaptable: Your code can handle new data types easily.\n",
    "\n",
    "Bad example:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "9bc290c8",
   "metadata": {},
   "outputs": [],
   "source": [
    "def analyze_data(data):\n",
    "    if isinstance(data, np.ndarray):\n",
    "        print(f\"Mean: {np.mean(data)}\")\n",
    "        print(f\"Standard Deviation: {np.std(data)}\")\n",
    "    elif isinstance(data, pd.Series):\n",
    "        print(f\"Mean: {data.mean()}\")\n",
    "        print(f\"Standard Deviation: {data.std()}\")\n",
    "    elif isinstance(data, CustomDataFrame):\n",
    "        print(f\"Mean: {data.mean()}\")\n",
    "        print(f\"Standard Deviation: {data.std()}\")\n",
    "    else:\n",
    "        raise TypeError(\"Unsupported data type\")\n",
    "\n",
    "\n",
    "# If you introduce a new data type, you have to modify the function\n",
    "class NewDataType:\n",
    "    def __init__(self, data):\n",
    "        self.data = data\n",
    "    def mean(self):\n",
    "        return sum(self.data) / len(self.data)\n",
    "    def std(self):\n",
    "        mean = self.mean()\n",
    "        return (sum((x - mean) ** 2 for x in self.data) / len(self.data)) ** 0.5\n",
    "\n",
    "new_data = NewDataType([1, 2, 3, 4, 5])\n",
    "# This will raise a TypeError\n",
    "# analyze_data(new_data)"
   ]
  }
 ],
 "metadata": {
  "celltoolbar": "Tags",
  "hide_input": false,
  "interpreter": {
   "hash": "484329849bb907480cd798e750759bc6f1d66c93f9e78e7055aa0a2c2de6b47b"
  },
  "kernelspec": {
   "display_name": "Python 3.8.9 ('venv': venv)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.6"
  },
  "toc": {
   "base_numbering": 1,
   "nav_menu": {},
   "number_sections": true,
   "sideBar": true,
   "skip_h1_title": false,
   "title_cell": "Table of Contents",
   "title_sidebar": "Contents",
   "toc_cell": false,
   "toc_position": {},
   "toc_section_display": true,
   "toc_window_display": false
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
